#include "Physics.h"

#include "Body.h"
#include <algorithm>

/** Whether the impulse solving should iterate multiple times, odd number! */
static constexpr int SimIterations = 1;

/** How often should the all body positions be evaluated per second. */
static constexpr float FpsTarget = 200.f;

/** Scaling the time per frame, thereby large values appear to slow time. */
static constexpr float TimeDilation = 1.f;

/**
 * @brief Start the physics thread
 * 
 * @param scene The scene description
 * 
 * Start the asynchronous physics thread.
 */
void Physics::startSimulation( std::shared_ptr<Scene> scene )
{
    m_scene = scene;
    m_thread = std::make_shared<std::thread>( &Physics::asyncTask, this );
}

/**
 * @brief Terminate the simulation and stop physics thread.
 */
void Physics::abortSimulation()
{
    m_abort = true;
    m_thread->join();
}

/**
 * @brief The physics thread routine
 * 
 * This routine is executed by the asynchronous physics thread. It contains
 * the physics initialization (here only masses) and velocity integration to
 * yield positional updates. Collision detection determines whether bodies
 * are touching or even overlapping. Overlaps are solved by singular
 * separation of the bodies, whereas non-receeding collision are solved by
 * applying frictionless normal impulses to both bodies (it is assumed that
 * all balls have the same mass).
 */
void Physics::asyncTask()
{
    // Update masses
    for (auto body : m_scene->bodies) {
        if (body->simulateDynamics) {
            body->m_mass = 0.165f;
            body->m_invMass = 1.f/body->m_mass;
        } else {
            body->m_mass = 1.f;
            body->m_invMass = 0.f;
            continue;
        }
    }

    auto frameTimePrev = std::chrono::high_resolution_clock::now();

    // ***********************
    // *** Simulation loop ***
    // ***********************

    while (!m_abort)
    {
        // Slow down the main loop if it is going too fast
        auto frameTime = std::chrono::high_resolution_clock::now();
        auto dt_local = std::chrono::duration<float>(frameTime - frameTimePrev).count();
        static constexpr float DtDesired = (1.f/FpsTarget) * (1.f/TimeDilation);
        while ((dt_local = std::chrono::duration<float>(frameTime - frameTimePrev).count()) < DtDesired) {
            std::this_thread::yield();
            frameTime = std::chrono::high_resolution_clock::now();
        }

        // Moving average of delta time
        dt = dt*0.9f + dt_local*0.1f;
        frameTimePrev = frameTime;

        // Integrate velocity
        bool all_static = true;
        for (int i=0; i<m_scene->bodies.size(); ++i) {
            // Ignore static bodies (table)
            auto body = m_scene->bodies[i];
            if (!body->simulateDynamics)
                continue;
            
            const auto original_pos = body->getPosition();
 
            /**
             * TODO: Animation
             * 
             * Implementieren Sie die Animation der Kugeln, dafuer muss die
             * Position der Kugeln angepasst werden (wir ignorieren Rotationen
             * hier komplett, da sich nur Kugeln bewegen und sich die Rotation
             * bei diesen nur sehr gering auf deren Position auswirkt).
             * Jede Kugel ist ein physikalischer Koerper, inklusive der Zustands-
             * variablen: Position und Geschwindigkeit. Die Geschwindigkeit wird
             * Beim Anstoß der Kugeln und bei Kollisionen direkt verändert. Die
             * Auswirkung der Geschwindigkeit auf die Position muss hier
             * implementiert werden. Dafür sind zwei Schritte notwendig:
             *
             * 1. Reibungsreduktion der Geschwindigkeit
             *   Implementieren Sie eine Reduktion der Geschwindigkeit in die
             *   umgekehrte Richtung der Geschwindigkeit. Dadurch wird ein
             *   einfaches Reibungsmodell zwischen Ball und Tischbezug realisiert.
             *   Innerhalb einer Sekunde geschieht ein Reibungsverlust der
             *   Geschwindigkeit vom Betrag Friction entgegen der Laufrichtung,
             *   jedoch nur sofern der Betrag der aktuellen Geschwindig-
             *   keit nicht größer als die Reduktion ist, ansonsten wird die
             *   Geschwindigkeit der Kugel auf 0 gesetzt.
             *   Diese Unterscheidung ist wichtig damit die Reibung niemals
             *   Energie in das System einführt, sondern stets Energie entnimmt,
             *   so dass die Billardbälle irgendwann auch zum Stillstand kommen.
             *
             * 2. Integration der Geschwindigkeit:
             *   Die gedämpfte Geschwindigkeit soll jetzt integriert werden um
             *   die Änderung der Position in diesem Zeitschritt zu erhalten:
             *                      p = p_old + v * dt
             *   Die z-Komponente (von Position & Geschwindigkeit soll jedoch
             *   stets auf Höhe des Tisches sein. Das ist eine weitere
             *   Vereinfachung des Modells.
             */   

            static constexpr float Friction = 0.7f;
            // TODO: Reibung
	    // TODO: Integration

            // Check if any ball moved considerably this frame
            const auto updated_pos = body->getPosition();
            static constexpr float VelocityEpsilon = 1.f*DtDesired;
            if (glm::length(updated_pos-original_pos)/dt > VelocityEpsilon)
                all_static = false;
        }
        m_ballsMoving = !all_static;

        struct Contact {
            std::shared_ptr<Body> body1;
            std::shared_ptr<Body> body2;
        };
        std::vector<Contact> contacts;

        // Detect collisions
        for (int i=0; i<m_scene->bodies.size(); ++i) {
            auto body = m_scene->bodies[i];
            if (!body->simulateDynamics)
                continue;

            // Ball + ball contacts
            for (int j=i+1; j<m_scene->bodies.size(); ++j) {
                auto otherBody = m_scene->bodies[j];
                if (!otherBody->simulateDynamics)
                    continue;
                auto deltaPosition = body->getPosition() - otherBody->getPosition();
                auto distance = glm::length(deltaPosition);
                auto overlap = (body->m_radius + otherBody->m_radius) - distance;
                if (overlap >= 0) {
                    Contact contact;
                    contact.body1 = body;
                    contact.body2 = otherBody;
                    contacts.push_back( contact );

                    // Separate overlapping balls
                    if (overlap > 0) {
                        auto impulseDirection = glm::normalize(deltaPosition);
                        impulseDirection.z = 0;
                        body->setPosition(body->getPosition() + impulseDirection * (overlap/2.f));
                        otherBody->setPosition(otherBody->getPosition() - impulseDirection * (overlap/2.f));
                    }
                }
            }
        }

        // Solve ball + ball contacts (velocity projection)
        for (int simiteration=0; simiteration<SimIterations; ++simiteration)
            for (auto contact : contacts)
                collisionBallBall(contact.body1, contact.body2);

        // Ball + bank
        for (int i=0; i<m_scene->bodies.size(); ++i) {
            auto body = m_scene->bodies[i];
            if (!body->simulateDynamics)
                continue;

            auto pos = body->getPosition();
            auto r = body->m_radius;
            static constexpr float Epsilon = 0.000001f;

            // Near + far
            static constexpr float BankY = 2.92305f+0.02f;
            if (-BankY-pos.y+r >= 0) {
                collisionBallBank( body, glm::vec3(0,+1,0) );
                pos.y = -BankY+r+Epsilon;
                body->setPosition( pos );
            }
            else if (BankY-pos.y-r <= 0) {
                collisionBallBank( body, glm::vec3(0,-1,0) );
                pos.y = +BankY-r-Epsilon;
                body->setPosition( pos );
            }

            // Left + right
            static constexpr float BankX = 1.41908f+0.01f;
            if (-BankX-pos.x+r >= 0) {
                collisionBallBank( body, glm::vec3(+1,0,0) );
                pos.x = -BankX+r+Epsilon;
                body->setPosition( pos );
            }
            else if (BankX-pos.x-r <= 0) {
                collisionBallBank( body, glm::vec3(-1,0,0) );
                pos.x = +BankX-r-Epsilon;
                body->setPosition( pos );
            }
        }
    }
}

/**
 * @brief Resolves a collision between two balls by applying impulses.
 * 
 * @param ball1 The first ball body.
 * @param ball2 The second ball body.
 * 
 * Assuming both balls have an equal mass, this function essentially
 * swaps out the normal components of the velocity of both balls.
 * The normal is determined by the line that connects the centers of
 * both balls. Additionally, the magnitude of the velocity is scaled
 * down by BallRestitution to simulate the energy loss due to friction.
 */
void Physics::collisionBallBall( std::shared_ptr<Body> ball1, std::shared_ptr<Body> ball2 )
{
    /**
     * TODO: Kollision Kugel + Kugel
     * 
     * Implementieren Sie die Kollisionsantwort für zwei Kugeln.
     * Die Geschwindigkeitsvektoren der Kugeln muss dafür nach der
     * Impulsgleichung vom Übungsblatt angepasst werden. Als
     * Elastizitätskoeffizient nehmen Sie BallResititution.
     */

    static constexpr float BallRestitution = 0.95f;

    // ball1->setVelocity( ... ); // Neuer Bewegungsvektor des ersten Balls
    // ball2->setVelocity( ... ); // Neuer Bewegungsvektor des zweiten Balls
}

/**
 * @brief Adjust velocity changes when a ball and a billard bank collide.
 * 
 * @param ball The ball body object pointer.
 * @param normal The normal that represents an infinite plane that aligns
 * with the colliding bank of the billiard table.
 * 
 * Assuming both balls have an equal mass, this function essentially
 * swaps out the normal components of the velocity of both balls.
 * The normal is determined by the line that connects the centers of
 * both balls. Additionally, the magnitude of the velocity is scaled
 * down by BallRestitution to simulate the energy loss due to friction.
 */
void Physics::collisionBallBank( std::shared_ptr<Body> ball, glm::vec3 normal )
{
    /**
     * TODO: Kollision Kugel + Bande
     * 
     * Implementieren Sie die Kollisionsantwort für Kugel und Bande.
     * Der Geschwindigkeitsvektor der Kugel muss dafür entlang der Bande
     * reflektiert werden. Ausserdem sollen 20% der Geschwindigkeit nach
     * der Kollision mit der Bande verloren gehen (was in Realität durch
     * Reibung und andere Effekte auch passieren würde).
     */

    static constexpr float BankRestitution = 0.8f;

    // ball->setVelocity(...); // Neuer Bewegungsvektor des Balls
}